from typing import *
from enum import Enum
import re

__all__ = []

RE_END_ANCHOR = re.compile(
    r'''
    (?:
        \(  # Group parentheses
            (?:  # Non-capturing, lookahead, named or flagged group specifiers
                \?
                (?:
                    [:=]
                |
                    P<[^>]+>
                |
                    (?: [aiLmsux]*(?: -[imsx]+\: ) )
                )
            )?
            (  # Actual group contents
                (?:  # Any non-escaped character class that may include parentheses or escaped character class brackets
                    (?<! \\ )
                    \[
                    (?: \\\] | [^\]] )*
                    (?<! \\ )\]
                |  # .. Or anything other than parentheses, character class brackets or escaped closing parentheses
                    (?: [^[\]()] | \\\) )*
                )*
            )
        \)
        (  # Spacing patterns between group and end of full expression
            (?:
                (?:  # Any spacing characters or the spacing character class
                    \\s | \s
                )
                (?:  # .. Possibly quantified or optional
                    [+*?]{1,2} | \{[0-9]*,?[0-9]*\}\??
                )?
            )*
            \$?
        )
        $
    )
    ''',
    re.VERBOSE
)


class MatcherType(Enum):
    EXACT = 1
    SUBSTRING = 2
    REGEXP = 3
    ANCHORED_REGEXP = 4


class PromptMatcher(object):
    pattern: Union[str, Tuple[str, str]] = None
    matcher_type: MatcherType = None

    def __init__(self, pattern: str, matcher_type: MatcherType = MatcherType.ANCHORED_REGEXP) -> None:
        self.pattern = pattern
        self.matcher_type = matcher_type
        getattr(self, '__init_' + matcher_type.name.lower() + '__')()
        setattr(self, 'matches', getattr(self, '_matches_' + self.matcher_type.name.lower()))

    def __init_exact__(self) -> None:
        pass

    def __init_substring__(self) -> None:
        pass

    def __init_regexp__(self) -> None:
        self.pattern = re.compile(self.pattern)

    def __init_anchored_regexp__(self) -> None:
        end_anchor = RE_END_ANCHOR.search(self.pattern)
        if end_anchor:
            end_anchor = end_anchor.group(1) + end_anchor.group(2)
            end_anchor += '' if end_anchor.endswith('$') else '$'
            self.pattern = (re.compile(self.pattern), re.compile(end_anchor))
        else:
            self.matcher_type = MatcherType.REGEXP
            self.__init_regexp__()

    def _matches_exact(self, test) -> bool:
        return test == self.pattern

    def _matches_substring(self, test) -> str:
        if self.pattern in test:
            return test

    def _matches_regexp(self, test) -> Type[re.Match]:
        return self.pattern.search(test)

    def _matches_anchored_regexp(self, test) -> Type[re.Match]:
        if self.pattern[1].search(test):
            return self.pattern[0].search(test)




from typing import *
from .io import ConsoleIO, SerialConsoleIO
from .matching import PromptMatcher, MatcherType
import io
import os
import time

__all__ = ['Console']


class Mapping(object):
    _from = None
    _from_str = None
    _from_len = 0
    _to = None
    _to_str = None
    _to_len = 0

    def __init__(self, map_from, map_to):
        self._from = list(map_from)
        self._from_str = map_from
        self._from_len = len(map_from)
        self._to = list(map_to)
        self._to_str = map_to
        self._to_len = len(map_to)

    def encode(self, text: str) -> str:
        return text.replace(self._to_str, self._from_str)

    def decode(self, text: str) -> str:
        return text.replace(self._from_str, self._to_str)

    def map_end(self, buffer: List[str], offset: int = 0) -> None:
        s = -self._from_len
        if offset > 0:
            s -= offset
            offset = -offset
        else:
            offset = None
        if buffer[s:offset] == self._from:
            buffer[s:offset] = self._to

    def map_all(self, buffer: List[str]) -> None:
        for i in range(len(buffer) - self._from_len):
            self.map_end(buffer, offset=i)

    def unmap_end(self, buffer: List[str], offset: int = 0) -> None:
        s = -self._to_len
        if offset > 0:
            s -= offset
            offset = -offset
        else:
            offset = None
        if buffer[s:offset] == self._to:
            buffer[s:offset] = self._from

    def unmap_all(self, buffer: List[str]) -> None:
        for i in range(len(buffer) - self._to_len):
            self.unmap_end(buffer, offset=i)


class Console(object):
    matcher: Type[PromptMatcher] = None
    mappings: List[Type[Mapping]] = []
    encoding: str = 'utf-8'
    linesep: str = list(os.linesep)
    unbuffered: bool = False
    timeout: float = -1
    strip_cmd: bool = True
    console_io: Type[ConsoleIO] = SerialConsoleIO
    output_io: Type[io.IOBase] = None

    command = None
    buffer = None
    prompt = None

    def __init__(self, pattern: Union[str, Type[PromptMatcher]],
                 pattern_type: MatcherType = MatcherType.ANCHORED_REGEXP,
                 mappings: List[Tuple[str, str]] = None,
                 encoding: str = encoding,
                 linesep: str = linesep,
                 unbuffered: bool = unbuffered,
                 timeout: float = timeout,
                 strip_cmd: bool = strip_cmd,
                 console_io_class: ClassVar[ConsoleIO] = console_io,
                 output_io: Type[io.IOBase] = output_io):
        if isinstance(pattern, PromptMatcher):
            self.matcher = pattern
        else:
            self.matcher = PromptMatcher(pattern=pattern, matcher_type=pattern_type)
        if mappings is not None:
            for mapping in mappings:
                self.mappings.append(Mapping(*mapping))
        self.encoding = encoding
        self.linesep = list(linesep)
        self.unbuffered = unbuffered
        self.timeout = timeout
        self.strip_cmd = strip_cmd
        self.console_io_class = console_io_class
        self.output_io = output_io

    def open(self, *args, **kwargs):
        io_class = self.console_io_class
        if 'io_class' in kwargs:
            io_class = kwargs['io_class']
            del kwargs['io_class']
        self.console_io = io_class(*args, **kwargs)
        self.console_io.open()

    def close(self):
        if self.console_io is not None:
            self.console_io.close()

    def send(self, command: Union[str, bytes], await_prompt: bool = True, **kwargs):
        if isinstance(command, str):
            for mapping in self.mappings:
                command = mapping.encode(command)
            self.command = command.rstrip()
            command = command.encode(self.encoding)
        self.console_io.write(command)
        if await_prompt:
            return self.await_prompt(**kwargs)

    def await_prompt(self,
                     matcher: Type[PromptMatcher] = None,
                     encoding: str = None,
                     unbuffered: bool = None,
                     timeout: float = None,
                     strip_cmd: bool = None,
                     output_io: Type[io.IOBase] = None) -> Union[None, str]:
        """
        Read from the console until the prompt is encountered.
        If `timeout` is non-negative, a TimeoutError will be raised if no new bytes have been received in `timeout` seconds.
        Returns the output buffer as a str by default if no arguments are provided.
        If `out_io` is set, then nothing is returned.
        Only writes to the output buffer once a linefeed is encountered (identified by `linesep`), unless `unbuffered`
        is True, in which case every character is written.
        Character mappings are not applied to the output IO in unbuffered mode, but are applied before matching for the prompt.
        The fully mapped output buffer is still available in the `buffer` class property.
        The `prompt` class property will be set to the matched prompt on successful matching.

        :param matcher: optional call override
        :param encoding: optional call override
        :param unbuffered: optional call override
        :param timeout: optional call override
        :param strip_cmd: optional call override
        :param output_io: optional call override
        :return: returns the output buffer as a str when `output_io` is None, otherwise None
        :raises TimeoutError: when `timeout` is non-negative and has been reached while waiting for the prompt
        """
        matcher = self.matcher if matcher is None else matcher
        encoding = self.encoding if encoding is None else encoding
        unbuffered = self.unbuffered if unbuffered is None else unbuffered
        timeout = self.timeout if timeout is None else timeout
        strip_cmd = self.strip_cmd if strip_cmd is None else strip_cmd
        output_io = self.output_io if output_io is None else output_io

        self.buffer = ""
        line_buffer = []
        line_count = 0
        prev_available = 0
        last_received_at = time.time()
        return_str = False
        lf_len = len(self.linesep)
        if output_io is None:
            return_str = True
            output_io = io.StringIO()

        while True:
            available = self.console_io.available()
            if available > 0:
                if available >= prev_available and timeout >= 0:
                    last_received_at = time.time()
                prev_available = available
                c = self.console_io.read(1).decode(encoding)
                line_buffer.append(c)
                if unbuffered:
                    output_io.write(c)
                for mapping in self.mappings:
                    mapping.map_end(line_buffer)
                match = matcher.matches(''.join(line_buffer))
                if match:
                    self.prompt = match
                    break
                if line_buffer[-lf_len:] == self.linesep:
                    line_count += 1
                    line_buffer = ''.join(line_buffer)
                    if not strip_cmd or line_count > 1 or line_buffer.strip() != self.command:
                        if not unbuffered:
                            output_io.write(line_buffer)
                        self.buffer += line_buffer
                    line_buffer = []
            else:
                if timeout >= 0 and (last_received_at + timeout) <= time.time():
                    raise TimeoutError("Timed out while waiting for prompt")
                time.sleep(0.005)

        if return_str:
            return output_io.getvalue()




from typing import *
from .console import Console
from .console.matching import PromptMatcher, MatcherType
from ._argparse import formatter
import argparse
import codecs
import sys
import re


def escaped_str(_str, enc=sys.getdefaultencoding()):
    return codecs.escape_decode(_str.encode(enc))[0].decode(enc)


def get_args():
    parser = argparse.ArgumentParser(
        description='Simple Console Reader is a small utility to send commands to a console over a'
                    'serial connection. It will continue to read until a prompt is detected, or it'
                    'times out (only if a timeout is specified).',
        epilog=['Since most prompts end with a special character (f.i. \'$\', \'#\', \'>\', etc.),'
                'a regular expression for the prompt may put the expression matching this '
                'character (or character group) inside a (non-) capturing group at the end of'
                'the full regular expression. This group will be detected and used to match'
                'first against a given output on the serial connection to speed up prompt'
                'matching.',
                'SCR supports normal capture groups \'(...)\', non-capture groups \'(?:...)\','
                'flag setting groups \'(?[aiLmsux]*-[imsx]+:...)\', and named groups'
                '\'(?P<[^>]+>...)\'. Beware that flag only groups \'(?[aiLmsux]+)\' will be'
                'interpreted as normal capture groups. Putting them at the end of the'
                'pattern will thus break the prompt matching.',
                'Example:',
                'The regular expression \'^[^#$]*(?:[#$]) $\' would start matching everything'
                'until it finds a \'#\' or \'$\' right at the end of a line. SCR will make use of'
                'the \'(?:[#$])\' group and match for \'[#$] $\' first (keeping any extra whitespace'
                'or \\s after it), which will drop any non-matches much faster ([#$] is rather'
                'uncommon compared to [^#$]).'],
        formatter_class=formatter.ParagraphedArgumentDefaultsHelpFormatter
    )
    if parser.prog == '__main__.py':
        parser.prog = 'python3 -m serial_console'
    parser.add_argument('-d', '--device', default='/dev/ttyS0', help='a serial device')
    parser.add_argument('-b', '--baud', type=int, default=9600, help='the baudrate of the serial device')
    parser.add_argument('-e', '--encoding', default='ascii', help='the encoding to communicate with')
    parser.add_argument('-t', '--timeout', type=float, default=None, help='a timeout to terminate after if no prompt could be detected')
    parser.add_argument('-m', '--map', type=escaped_str, metavar=('CONSOLE', 'LOCAL'), action='append', nargs=2,
                        help=['two sequences which should be mapped during communication from/to the serial device.',
                              'May be specified multiple times.',
                              'Escape sequences will be replaced (i.e. \\n becomes a literal newline).',
                              'LOCAL may be an empty string.'])
    parser.add_argument('-u', '--unbuffered', action='store_true', help='')
    parser.add_argument('-g', '--get-prompt', action='store_true', help='print the parsed prompt after each command execution')
    parser.add_argument('commands', type=escaped_str, metavar='CMD', nargs='+', help='one or multiple commands to send to the console')
    prompt_group = parser.add_mutually_exclusive_group()
    prompt_group.add_argument('-r', '--regex-prompt', metavar='REGEXP', help='a regular expression matching the prompt')
    prompt_group.add_argument('-s', '--substr-prompt', metavar='SUBSTR', help='a substring matching the prompt')
    prompt_group.add_argument('-p', '--prompt', help='a string matching the prompt')

    return parser.parse_args()


def get_console_args(args: argparse.Namespace):
    pattern = args.prompt
    pattern_type = MatcherType.EXACT
    if args.regex_prompt is not None:
        pattern_type = MatcherType.ANCHORED_REGEXP
        pattern = args.regex_prompt
    if args.substr_prompt is not None:
        pattern_type = MatcherType.SUBSTRING
        pattern = args.substr_prompt

    return [PromptMatcher(pattern, matcher_type=pattern_type)]


def get_console_kwargs(args: argparse.Namespace):
    kwargs = {
        'mappings': args.map,
        'encoding': args.encoding,
        'unbuffered': args.unbuffered
    }

    if args.timeout is not None:
        kwargs['timeout'] = args.timeout

    return kwargs


def get_console_open_args(args: argparse.Namespace):
    return [args.device, args.baud]


def print_prompt(prompt: Union[str, Type[re.Match]]):
    if isinstance(prompt, re.Match):
        def p(k, v): print('{}\t{}'.format(k, v))
        p(0, prompt.group(0))
        for i, grp in enumerate(prompt.groups(), 1):
            p(i, grp)
        for key, grp in prompt.groupdict():
            p(key, grp)
    else:
        print(prompt)


def main():
    status = 0
    args = get_args()
    cli = Console(*get_console_args(args), strip_cmd=True, output_io=sys.stdout, **get_console_kwargs(args))
    cli.open(*get_console_open_args(args))
    for cmd in args.commands:
        try:
            cli.send(cmd)
            if args.get_prompt:
                print('\0', end='')
                print_prompt(cli.prompt)
        except TimeoutError:
            status = 1
            print('\x1b[31mCommand timed out:\x1b[0m \'{}\''.format(cmd), file=sys.stderr)
            print('\x1b[31mBuffer:\x1b[0m', file=sys.stderr)
            print(cli.buffer, file=sys.stderr)

    sys.exit(status)



import argparse
import copy

__all__ = ['ParagraphedDescriptionHelpFormatter', 'ParagraphedArgumentDefaultsHelpFormatter']


class ParagraphedDescriptionHelpFormatter(argparse.HelpFormatter):
    @staticmethod
    def _is_iterable(obj):
        return isinstance(obj, list) or isinstance(obj, tuple) or isinstance(obj, set) or isinstance(obj, frozenset)

    def _format_text(self, text):
        if self._is_iterable(text):
            paragraphs = []
            for paragraph in text:
                paragraphs.append(super()._format_text(paragraph)[0:-2])
            return "\n".join(paragraphs) + "\n\n"
        else:
            return super()._format_text(text)

    def _split_lines(self, text, width):
        if self._is_iterable(text):
            paragraphs = []
            for paragraph in text:
                paragraphs.extend(super()._split_lines(paragraph, width))
            return paragraphs
        else:
            return super()._split_lines(text, width)

    def _expand_help(self, action):
        if self._is_iterable(action.help):
            action_copy = copy.deepcopy(action)
            paragraphs = []
            for paragraph in action.help:
                action_copy.help = paragraph
                paragraphs.append(super()._expand_help(action_copy))
            return paragraphs
        else:
            return super()._expand_help(action)


class ParagraphedArgumentDefaultsHelpFormatter(ParagraphedDescriptionHelpFormatter):
    suppress_none_defaults = True

    def _get_defaults_string(self, action):
        if action.default is not argparse.SUPPRESS and (self.suppress_none_defaults is False or action.default is not None):
            defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
            if action.option_strings or action.nargs in defaulting_nargs:
                return "(default: %(default)s)"
        return None

    def _expand_help(self, action):
        default_string = self._get_defaults_string(action)
        if default_string is not None:
            if self._is_iterable(action.help):
                has_default = False
                paragraphs = []
                for paragraph in action.help:
                    paragraphs.append(paragraph)
                    if '%(default)' in paragraph:
                        has_default = True
                if not has_default:
                    action_copy = copy.deepcopy(action)
                    action_copy.help = paragraphs + [default_string]
                    return super()._expand_help(action_copy)
            else:
                action_copy = copy.deepcopy(action)
                action_copy.help += " " + default_string
                return super()._expand_help(action_copy)
        return super()._expand_help(action)




from setuptools import setup, find_namespace_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="hlfbt-serial-console",
    version="0.0.2",
    author="Alexander Schulz",
    author_email="alex@nope.bz",
    description="A simple utility to interface with prompt-like (serial) consoles",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hlfbt/serial_console",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Communications",
        "Topic :: Terminals :: Serial"
    ],
    python_requires='>=3.5',
    packages=find_namespace_packages(include=['hlfbt.*']),
    install_requires=['pyserial']
)



import serial

__all__ = []


class ConsoleIO(object):

    def open(self, *args, **kwargs):
        """
        Prepare and open the connection.
        """
        NotImplementedError("Class %s doesn't implement aMethod()" % self.__class__.__name__)

    def close(self):
        """
        Terminate and close the connection.
        """
        NotImplementedError("Class %s doesn't implement aMethod()" % self.__class__.__name__)

    def read(self, num: int = 1) -> bytes:
        """
        Read `num` number of bytes from the read buffer.

        :param num: the number of bytes to read
        :return: the retrieved bytes
        """
        NotImplementedError("Class %s doesn't implement aMethod()" % self.__class__.__name__)

    def write(self, buffer: bytes) -> None:
        """
        Write `buffer` to the write buffer.

        :param buffer: the bytes to write
        """
        NotImplementedError("Class %s doesn't implement aMethod()" % self.__class__.__name__)

    def flush(self) -> None:
        """
        Flush all remaining bytes from the write buffer.
        Wait until all data is written.
        """
        NotImplementedError("Class %s doesn't implement aMethod()" % self.__class__.__name__)

    def available(self):
        """
        Get the number of bytes available in the read buffer.

        :return: the number of available bytes
        """
        NotImplementedError("Class %s doesn't implement aMethod()" % self.__class__.__name__)


class SerialConsoleIO(serial.Serial, ConsoleIO):

    def available(self):
        return self.in_waiting

    def open(self):
        if self.is_open:
            return
        return super().open()

